/* checkEnds - Check that ends output for a set is reasonable. */
#include "common.h"
#include "linefile.h"
#include "hash.h"
#include "dlist.h"

boolean anyProblem = FALSE;

void report(char *format, ...)
/* Print error message to stdout and note there was a problem. */
{
va_list args;
va_start(args, format);
vfprintf(stdout, format, args);
va_end(args);
anyProblem = TRUE;
}


void usage()
/* Explain usage and exit. */
{
errAbort(
  "checkEnds - Check that ends output for a set is reasonable\n"
  "usage:\n"
  "   checkEnds testSet\n");
}

struct clone
/* Info on a clone. */
    {
    struct clone *next;
    char *name;
    int start;
    int end;
    int orientation;
    struct cloneEnd *endList;
    };

struct cloneEnd
/* Info on a clone end. */
    {
    struct cloneEnd *next;
    boolean isStart;
    struct clone *clone;
    };

void parseCloneEnd(char *in, char **retClone, int *retStart, int *retEnd,
	int *retOrientation, boolean *retIsStart)
/* Parse out a clone end string into fields. */
{
static char name[64];
char dupe[64], *parts[4];
int start, end, partCount;
int orientation = 0;
boolean isStart = FALSE;
char c;
char *s = dupe;

strcpy(dupe, in);
c = *s;
if (c == '+')
    {
    isStart = TRUE;
    orientation = +1;
    s += 2;
    }
else if (c == '-')
    {
    isStart = TRUE;
    orientation = -1;
    s += 2;
    }
else if (c == '?')
    {
    isStart = TRUE;
    orientation = 0;
    s += 2;
    }
partCount = chopString(s, "@_)", parts, ArraySize(parts));
if (partCount != 3)
    errAbort("Unparsable clone name %s", in);
strcpy(name, parts[0]);

if (!isdigit(parts[1][0]) || !isdigit(parts[2][0]))
    errAbort("Expecting integer in clone name %s", in);
start = atoi(parts[1]);
end = atoi(parts[2]);

*retClone = name;
*retStart = start;
*retEnd = end;
*retIsStart = isStart;
*retOrientation = orientation;
}

int countForwardEnds(struct lineFile *lf, struct dlList *endList, boolean squawk)
/* Count up number of ends in a row that could be forward. */
{
struct dlNode *node;
int lastPos = -1, pos;
struct cloneEnd *ce;
struct clone *clone;
int count = 0;

for (node = endList->head; !dlEnd(node); node = node->next)
    {
    ce = node->val;
    clone = ce->clone;
    if (ce->isStart)
        pos = clone->start;
    else
        pos = clone->end;
    if (lastPos <= pos)
	++count;
    else 
	{
        if (squawk)
	    report("Reversal at clone %s line %d of %s\n", 
	        clone->name, lf->lineIx, lf->fileName);
	break;
	}
    lastPos = pos;
    }
return count;
}


int countReverseEnds(struct lineFile *lf, struct dlList *endList, boolean squawk)
/* Count up number of ends in a row that could be reverse. */
{
struct dlNode *node;
int lastPos = -1, pos;
struct cloneEnd *ce;
struct clone *clone;
int count = 0;

for (node = endList->tail; !dlStart(node); node = node->prev)
    {
    ce = node->val;
    clone = ce->clone;
    if (ce->isStart)
        pos = clone->end;
    else
        pos = clone->start;
    if (lastPos <= pos)
	++count;
    else 
	{
        if (squawk)
	    report("Backwards reversal at clone %s line %d of %s\n", 
	        clone->name, lf->lineIx, lf->fileName);
	break;
	}
    lastPos = pos;
    }
return count;
}


void checkBarge(struct lineFile *lf, int endCount, char *endStrings[])
/* Make sure that barge is internally consistent. */
{
struct hash *cloneHash = newHash(8);
char *cloneName;
int i,start,end,orientation;
boolean isStart;
char *endString;
struct clone *cloneList = NULL, *clone;
struct cloneEnd *ce, *lastCe = NULL;
int lastStart, lastEnd;
int lastPos, lastRevPos, pos, revPos;
int direction = 0;
struct dlList *endList = newDlList();
int forCount, revCount;

if (endCount&1)
    {
    report("Odd clone end count line %d of %s\n", lf->lineIx, lf->fileName);
    return;
    }

for (i=0; i<endCount; ++i)
    {
    endString = endStrings[i];
    parseCloneEnd(endString, &cloneName, &start, &end, &orientation, &isStart);
    clone = hashFindVal(cloneHash, cloneName);
    if (clone == NULL)
       {
       if (!isStart)
	   {
	   report("First appearance of clone %s is not a start line %d of %s\n",
		endString, lf->lineIx, lf->fileName);
	   }
       AllocVar(clone);
       hashAddSaveName(cloneHash, cloneName, clone, &clone->name);
       clone->start = start;
       clone->end = end;
       clone->orientation = orientation;
       slAddHead(&cloneList, clone);
       }
    else
       {
       if (isStart)
	   {
           report("Second appearance of clone %s is not end line %d of %s\n",
		endString, lf->lineIx, lf->fileName);
	   }
	       
       }
    AllocVar(ce);
    ce->isStart = isStart;
    ce->clone = clone;
    slAddTail(&clone->endList, ce);
    dlAddValTail(endList, ce);
    }

forCount = countForwardEnds(lf, endList, FALSE);
revCount = countReverseEnds(lf, endList, FALSE);
if (forCount != endCount && revCount != endCount)
    {
    if (forCount >= revCount)
        countForwardEnds(lf, endList, TRUE);
    else
        countReverseEnds(lf, endList, FALSE);
    }

for (clone = cloneList; clone != NULL; clone = clone->next)
    {
    int endCount = slCount(clone->endList);
    if (endCount != 2)
        report("Got %d ends on clone %s line %d of %s\n", endCount, clone->name,
	    lf->lineIx, lf->fileName);
    }
freeDlListAndVals(&endList);
slFreeList(&cloneList);
freeHash(&cloneHash);
}

void checkEndsFile(char *fileName)
/* Check consistency of ends in one file. */
{
struct lineFile *lf = lineFileOpen(fileName, TRUE);
static char *words[4096], *line;
int wordCount, lineSize;

while (lineFileNext(lf, &line, &lineSize))
    {
    /* Skip blank and commented lines. */
    if (line[0] == '#')
        continue;
    wordCount = chopLine(line, words);
    if (wordCount == 0)
        continue;
    checkBarge(lf, wordCount, words);
    }
lineFileClose(&lf);
}

void checkEnds(char *testSet)
/* checkEnds - Check that ends output for a set is reasonable. */
{
char fileName[512];
sprintf(fileName, "%s.ends", testSet);
checkEndsFile(fileName);
if (!anyProblem)
    printf("No problems detected in %s\n", fileName);
}

int main(int argc, char *argv[])
/* Process command line. */
{
if (argc != 2)
    usage();
checkEnds(argv[1]);
return 0;
}
